/*global FileManager, log, module, _, utils:true, converter:true */

var pluginInfo = { "Name":"Nordic Semiconductor nRF5x v.1.2.4", "Description":"Generates header and code files for nRF5x.", "Author":"Nordic Semiconductor", "Version":"1.2.4", "IsClient":false };

function IncludeScript(filename) {
    var include = FileManager.ReadFile(filename);
    /*jslint evil: true */
    var newfunc = new Function(include);
    var myfunk = newfunc();
}

function GetInfo(infoObject) {
    infoObject.Name = pluginInfo.Name;
    infoObject.Description = pluginInfo.Description;
    infoObject.Author = pluginInfo.Author;
    infoObject.Version = pluginInfo.Version;
    infoObject.IsClient = pluginInfo.IsClient;

    return infoObject;
}

function generateFiles(jsondata) {
    var data;
    pluginInfo.AppVersion = jsondata.AppVersion;
    jsondata.PluginInfo = pluginInfo;

    // Temporarily store service short names (used to check for uniqueness of service within profile)
    var serviceNames = [];
    _.each(jsondata.Services, function (service) {
        service.ShortName = getServiceShortName(service.Name);
        serviceNames.push(service.ShortName);
    });
    var uniqueServiceNames = _.countBy(serviceNames, function(name) { return name; });

    // Enrich services and generate header/code files
    _.each(jsondata.Services, function (service) {
        service.IsUnique = (uniqueServiceNames[service.ShortName] === 1) ? true : false;
        enrich(service);
        log("Generating files for service " + service.Name + "...");
        data = createHeaderFile(service);
        FileManager.CreateFile("ble_" + service.ShortName + ".h", data);
        data = createCodeFile(service);
        FileManager.CreateFile("ble_" + service.ShortName + ".c", data);
    });

    // Get unique service names (possibly with UUIDs)
    uniqueServiceNames = _.countBy(jsondata.Services, function(service) { return service.ShortName; });

    // Append counter value to differentiate variable names of identical services within profile, if needed
    for(var i=jsondata.Services.length-1; i>=0; i--) {
        var shortName = jsondata.Services[i].ShortName;
        if (uniqueServiceNames[shortName] > 1) {
            jsondata.Services[i].UniqueName = shortName + "_" + uniqueServiceNames[shortName];
        }
        else {
            jsondata.Services[i].UniqueName = shortName;
        }
        uniqueServiceNames[shortName]--;
    }

    // Generate template-to-generated-code interface files
    log("Generating SDK template project interface files...");
    data = createInterfaceHeaderFile();
    FileManager.CreateFile("service_if.h", data);

    data = createInterfaceCodeFile(jsondata);
    FileManager.CreateFile("service_if.c", data);
}

function RunPlugin(profiledata) {
    converter = {};
    IncludeScript("json_converter.js");
    IncludeScript("underscore_customized.js");

    log("Starting code generation...");
    var jsondata = converter.ConvertToJson(profiledata);

    generateFiles(jsondata);
    var readmeFile = FileManager.ReadFile("Readme.txt");
    FileManager.CreateFile("Readme.txt", readmeFile);
    log("Code generation complete.");
}

function enrich(service) {
    var baseExtras = {
        NormalizedName: function () {
            return utils.getNormalizedName(this.Name);
        }
    };

    var characteristicsExtras = _.extend({}, baseExtras, {
        isReadSupported: function () {
            return (this.Properties[0].Read !== PropertyRequirementEnum.Excluded);
        },
        isReadMandatory: function () {
            return (this.Properties[0].Read === PropertyRequirementEnum.Mandatory);
        },
        isNotifySupported: function () {
            return (this.Properties[0].Notify !== PropertyRequirementEnum.Excluded);
        },
        isNotifyMandatory: function () {
            return (this.Properties[0].Notify === PropertyRequirementEnum.Mandatory);
        },
        isIndicateSupported: function () {
            return (this.Properties[0].Indicate !== PropertyRequirementEnum.Excluded);
        },
        isIndicateMandatory: function () {
            return (this.Properties[0].Indicate === PropertyRequirementEnum.Mandatory);
        },
        isWriteSupported: function () {
            return (this.Properties[0].Write !== PropertyRequirementEnum.Excluded);
        },
        isWriteMandatory: function () {
            return (this.Properties[0].Write === PropertyRequirementEnum.Mandatory);
        },
        isWriteWithoutResponseSupported: function () {
            return (this.Properties[0].WriteWithoutResponse !== PropertyRequirementEnum.Excluded);
        },
        isWriteWithoutResponseMandatory: function () {
            return (this.Properties[0].WriteWithoutResponse === PropertyRequirementEnum.Mandatory);
        },
        isReliableWriteSupported: function () {
            return (this.Properties[0].ReliableWrite !== PropertyRequirementEnum.Excluded);
        },
        isReliableWriteMandatory: function () {
            return (this.Properties[0].ReliableWrite === PropertyRequirementEnum.Mandatory);
        },
        SendFunction: function () {
            // Check if we need to generate a "send" function
            if (this.isNotifySupported() && this.isIndicateSupported()) {
                return [
                    {
                        Postfix: 'notify',
                        Type: 'BLE_GATT_HVX_NOTIFICATION'
                    },
                    {
                        Postfix: 'indicate',
                        Type: 'BLE_GATT_HVX_INDICATION'
                    }
                ];
            } else if (this.isNotifySupported() || this.isIndicateSupported()) {
                return [
                    {
                        Postfix: 'send',
                        Type: this.isNotifySupported() ? 'BLE_GATT_HVX_NOTIFICATION' : 'BLE_GATT_HVX_INDICATION'
                    }
                ];
            }
        },
        StructName: function () {
            var structname;
            if (this.Fields.length > 0) {
                structname = "ble_" + service.ShortName.toLowerCase() + "_" + this.NormalizedName() + "_t";
            }
            return structname;
        },
        getFullUuid: function() {
            return utils.getUuidBase(this.UUID);
        },
        hasNibbles: function() {
            return _.some(this.Fields, function(field) {
                return field.FormatType() === 'nibble';
            });
        }
    });

    var descriptorExtras = _.extend({}, baseExtras, {
        isReadSupported: function () {
            return (this.Properties.Read !== PropertyRequirementEnum.Excluded);
        },
        isReadMandatory: function () {
            return (this.Properties.Read === PropertyRequirementEnum.Mandatory);
        },
        isWriteSupported: function () {
            return (this.Properties.Write !== PropertyRequirementEnum.Excluded);
        },
        isWriteMandatory: function () {
            return (this.Properties.Write === PropertyRequirementEnum.Mandatory);
        },
        StructName: function () {
            var structname;
            if (this.Fields.length > 0) {
                structname = "ble_" + service.ShortName.toLowerCase() + "_" + this.NormalizedName() + "_t";
            }
            return structname;
        },
        hasNibbles: function() {
            return _.some(this.Fields, function(field) {
                return field.FormatType() === 'nibble';
            });
        }
    });

    var enumList = [];
    var fieldExtras = _.extend({}, baseExtras, {
        getEnumList: function(){
            return (enumList);
        },
        isStruct: function (fieldName) {
            return ((this.Enumerations && this.getEnumerations(fieldName).length > 0) || (this.BitField && this.BitField.Bits.length > 0));
        },
        hasRequirement: function () {
            return !_.some(this.Requirement,function(element) {
                var liste = ["Mandatory", "Optional", "Conditional"];
                return _.contains(liste, element);
            });
        },
        FormatType: function() {
            return utils.getTypeDef(this.Format, this.Reference);
        },
        isVariableLength: function() {
            var var_length_types = ['uint8_array'];
            return _.contains(var_length_types, this.FormatType());
        },
        isEnumValid: function(fieldName){
            if (enumList.length === 0){
                enumList.push(fieldName);
                return true;
            }
            if (_.contains(enumList, fieldName)){
                return false;
            }
            var newList =[fieldName];
            enumList.push(newList);
            return true;
        },
        getEnumerations: function() {
            if (!_.isNull(this.Enumerations)) {
                return utils.resolveEnumDuplicates(this.Enumerations.Enumeration);
            }
        }
    });


    var enumBitList = [];
    var bitfieldExtras = _.extend({}, baseExtras, {
        getEnumList: function(){
            return (enumBitList);
        },
        isBitEnumValid: function(fieldName, bitfieldName){
            var enumName = fieldName + "_" + bitfieldName;
            var newList =[enumName];
            if (enumBitList.length === 0){
                enumBitList.push(enumName);
                return true;
            }
            if (_.contains(enumBitList, enumName)){
                return false;
            }
            enumBitList.push(enumName);
            return true;
        },
        getEnumerations: function() {
            if (!_.isNull(this.Enumerations)) {
                return utils.resolveEnumDuplicates(this.Enumerations.Enumeration);
            }
        }
    });

    var enumerationExtras = _.extend({}, baseExtras, {
        NormalizedName: function () {
            return utils.getNormalizedName(this.Value);
        },
        hasRequires: function () {
            return !_.isUndefined(this.Requires) && this.Requires !== "NA";
        }
    });

    _.extend(service, baseExtras);
    if(!service.IsUnique) {
        service.ShortName = getServiceShortName(service.Name) + '_' + utils.getNormalizedName(utils.getShortUuid(service.UUID));
    }
    service.FullUuid = utils.getUuidBase(service.UUID);
    service.PluginInfo = pluginInfo;

    _.each(service.Characteristics, function (characteristic) {
        _.extend(characteristic, characteristicsExtras);

        var indexOfFirstMatch = -1;
        _.each(characteristic.Descriptors, function (descriptor) { // Descriptors
            _.extend(descriptor, descriptorExtras);

            // Find the first descriptor which matches below UUID, tag it and store its index in the Descriptors array
            if(indexOfFirstMatch < 0) { // No match yet
                var firstMatch = _.find(characteristic.Descriptors, function (descriptor) {
                    return (descriptor.UUID === "2901" || descriptor.UUID === "2904");
                });

                if(!_.isUndefined(firstMatch)) {
                    indexOfFirstMatch = _.indexOf(characteristic.Descriptors, firstMatch);
                    descriptor.DescriptorAddSpecial = true;
                    characteristic.HasDescriptorAddSpecial = true;
                }
            }

            _.each(descriptor.Fields, function (field) { // Structs
                _.extend(field, fieldExtras);
                if (!_.isNull(field.Enumerations)) {
                    _.each(field.Enumerations.Enumeration, function (enumVal) {
                        _.extend(enumVal, enumerationExtras);
                    });
                }

                if(!_.isNull(field.BitField)) {
                    _.each(field.BitField.Bits, function (bitField) { // Enums
                        if (_.isNull(bitField.Name)) { // Hack to avoid null ref exception because of descriptor 0x2902 bit with null Name
                            bitField.Name =  service.Name + "_" + field.Name + "_bit_" + bitField.Index;
                        }
                        _.extend(bitField, bitfieldExtras);

                        // Enum values/members
                        if (!_.isNull(bitField.Enumerations)) {
                            _.each(bitField.Enumerations.Enumeration, function (enumVal) {
                                _.extend(enumVal, enumerationExtras);
                            });
                        }
                    });
                }
            });
        });

        // If a "special" descriptor was found (and it's not already at index 0), move it
        // to the beginning of the Descriptors array so it can be added first in code template
        if(indexOfFirstMatch > 0) {
            var removedItem = characteristic.Descriptors.splice(indexOfFirstMatch, 1);
            characteristic.Descriptors.unshift(removedItem[0]);
        }

        _.each(characteristic.Fields, function (field) { // Structs
            _.extend(field, fieldExtras);
            if (!_.isNull(field.Enumerations)) {
                _.each(field.Enumerations.Enumeration, function (enumVal) {
                    _.extend(enumVal, enumerationExtras);
                });
            }

            if(!_.isNull(field.BitField)) {
                _.each(field.BitField.Bits, function (bitField) { // Enums
                    if (_.isNull(bitField.Name)) { // Hack to avoid null ref exception because of descriptor 0x2902 bit with null Name
                        bitField.Name =  service.Name + "_" + field.Name + "_bit_" + bitField.Index;
                    }
                    _.extend(bitField, bitfieldExtras);

                    // Enum values/members
                    if (!_.isNull(bitField.Enumerations)) {
                        _.each(bitField.Enumerations.Enumeration, function (enumVal) {
                            _.extend(enumVal, enumerationExtras);
                        });
                    }
                });
            }
        });
    });
}


/**
 * Creates header file for the provided service
 * @param {JSON service object} service
 * @return {string} headerFile
 */
function createHeaderFile(service) {
    log("Generating header file ble_" + service.ShortName + ".h");
    var template = FileManager.ReadFile('header_template.h');
    var compiled = _.template(template);
    var data = compiled(service);
    return data;
}

/**
 * Creates code file for the provided service
 * @param {JSON service object} service
 * @return {string} codeFile
 */
function createCodeFile(service) {
    log("Generating code file ble_" + service.ShortName + ".c");
    var template = FileManager.ReadFile('code_template.c');
    var compiled = _.template(template);
    var data = compiled(service);
    return data;
}

/**
 * Creates header file for template-to-generated-code interface (static file)
 * @return {string} headerFile
 */
function createInterfaceHeaderFile() {
    log("Generating header file service_if.h");
    var template = FileManager.ReadFile('service_if_template.h');
    var compiled = _.template(template);
    var data = compiled();
    return data;
}

/**
 * Creates code file for all services in the provided profile (template-to-generated-code interface)
 * @param {JSON profile object} profile
 * @return {string} codeFile
 */
function createInterfaceCodeFile(profile) {
    log("Generating code file service_if.c");
    var template = FileManager.ReadFile('service_if_template.c');
    var compiled = _.template(template);
    var data = compiled(profile);
    return data;
}

/**
 * Returns lowercase three/four-letter short name for BLE profiles
 * See https://www.bluetooth.org/en-us/specification/adopted-specifications
 * @param {string} serviceName
 * @return {string} shortName
 */
function getServiceShortName(serviceName) {

    // TODO: Get this info from BDS service profile (SIG needs to add "shortname" tag)

    var shortName = "";
    switch (serviceName) {
        case "Alert Notification":
            shortName = "ans";
            break;
        case "Battery Service":
            shortName = "bas";
            break;
        case "Body Composition":
            shortName = "bcs";
            break;
        case "Blood Pressure":
            shortName = "bls";
            break;
        case "Continuous Glucose Monitoring":
            shortName = "cgms";
            break;
        case "Cycling Power":
            shortName = "cps";
            break;
        case "Cycling Speed and Cadence":
            shortName = "cscs";
            break;
        case "Current Time":
            shortName = "cts";
            break;
        case "Device Information":
            shortName = "dis";
            break;
        case "Environmental Sensing":
            shortName = "ess";
            break;
        case "Glucose":
            shortName = "gls";
            break;
        case "HID":
            shortName = "hids";
            break;
        case "Heart Rate":
            shortName = "hrs";
            break;
        case "Health Thermometer":
            shortName = "hts";
            break;
        case "Immediate Alert":
            shortName = "ias";
            break;
        case "Link Loss":
            shortName = "lls";
            break;
        case "Location and Navigation":
            shortName = "lns";
            break;
        case "Next DST Change":
            shortName = "ndcs";
            break;
        case "Phone Alert Status":
            shortName = "pass";
            break;
        case "Running Speed and Cadence":
            shortName = "rscs";
            break;
        case "Reference Time Update":
            shortName = "rtus";
            break;
        case "Scan Parameters":
            shortName = "scps";
            break;
        case "Tx Power":
            shortName = "tps";
            break;
        case "User Data":
            shortName = "uds";
            break;
        case "Weight Scale":
            shortName = "wss";
            break;
        default:
            shortName = utils.getNormalizedName(serviceName);
    }
    return shortName.toLowerCase();
}

utils = typeof utils !== 'undefined' ? utils : {};
(function (ns) {
    ns.getNormalizedName = function (name) {
        return name.toLowerCase().replace(/#|-| |:|\.|'|’|,|\(|\)|\//g, '_');
    };

    ns.getUuidAsArray = function (uuid) {
        var re = /([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})-([0-9a-f]{2})([0-9a-f]{2})-([0-9a-f]{2})([0-9a-f]{2})-([0-9a-f]{2})([0-9a-f]{2})-([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})/i;
        var arr = re.exec(uuid);
        if (_.isNull( !_.isNull(arr) ? arr.reverse() : null)){
            var re2 = /([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})([0-9a-f]{2})/i;
            var arr2 = re2.exec(uuid);
            return !_.isNull(arr2) ? arr2.reverse() : null;
        } else {
            return !_.isNull(arr) ? arr.reverse() : null;
        }
    };

    ns.getUuidBase = function(uuid) {
        var arr = ns.getUuidAsArray(uuid);
        if (_.isNull(arr)) {
            return null;
        } else {
            var uuidBase = '0x' + arr[0];
            for (var i = 1; i < 16; i++) {

                if (i === 12 || i === 13) {
                    uuidBase += ', 0x00';
                } else {
                    uuidBase += ', 0x' + arr[i];
                }
            }
            var str = '0x' + arr[13] + arr[12];
            return {'UuidBase': uuidBase, 'Uuid' : str};
        }
    };

    ns.getShortUuid = function(uuid) {
        if((uuid.length === 36)||((uuid.length === 32))) {
            return uuid.substring(4,4);
        }
        return uuid;
    };

    ns.resolveEnumDuplicates = function(enumerations) {
        var uniqueServiceNames = _.countBy(enumerations, function (enumVal) { return enumVal.Value; });
        _.each(enumerations, function (enumVal) {
            // Check for identical enum member names, append Key value to name if needed
            if (uniqueServiceNames[enumVal.Value] > 1) {
                enumVal.Value += '_' + enumVal.Key;
            }
        });
        return enumerations;
    };

    ns.getTypeDef = function (name, ref) {
        if (name === '2bit') {
            return 'uint8_array';
        }
        if (name === '4bit') {
            return 'nibble';
        }
        if (name === 'Item8bit' || name === '8bit') {
            return 'uint8';
        }
        if (name === 'Item16bit' || name === '16bit') {
            return 'uint16';
        }
        if (name === '24bit') {
            return 'uint24';
        }
        if (name === '32bit') {
            return 'uint32';
        }
        if (name === 'uint8[]') {
            return 'uint8_array';
        }
        if (name === 'uint12' || name === 'uint48' || name === 'uint128') {
            return 'uint8_array';
        }
        if (name === 'sint12' || name === 'sint24' || name === 'sint48' || name === 'sint128') {
            return 'uint8_array';
        }
        if (name === 'sint8') {
            return 'int8';
        }
        if (name === 'sint16') {
            return 'int16';
        }
        if (name === 'sint32') {
            return 'int32';
        }
        if (name === 'sint64') {
            return 'int64';
        }
        if (name === 'float32') {
            return 'float';
        }
        if (name === 'float64') {
            return 'double';
        }
        if (name === 'duint16') {
            return 'uint8';
        }
        if (name === 'utf8s') {
            return 'ble_srv_utf8_str';
        }
        if (name === 'utf16s') {
            return 'uint8_array';
        }
        if (name === 'characteristic' || name === 'struct' ||
            name === 'gatt_uuid' || name === 'variable') {
            return 'uint8_array';
        }
        if (name === 'reg-cert-data-list') {
            return 'regcertdatalist';
        }
        if (ref === 'org.bluetooth.characteristic.date_time'){
            return 'ble_date_time';
        }
        return (name !== null) ? name.toLowerCase() : 'UNDEFINED_TYPE';
    };

}(utils));

// "Enums"

// Bluetooth.D3T.Model.PropertyRequirements
// Used by Characteristic.Properties.Read/Write/...
// BDS only gives numerical values for these.
var PropertyRequirementEnum = Object.freeze({
    "Mandatory": "Mandatory",
    "Optional": "Optional",
    "Excluded": "Excluded",
    "C1": "C1",
    "C2": "C2",
    "C3": "C3",
    "C4": "C4",
    "C5": "C5",
    "C6": "C6",
    "C7": "C7",
    "C8": "C8",
    "C9": "C9",
    "C10": "C10"
});

// Bluetooth.D3T.Model.requirement
// Used by Characteristic field "Requirement"
// BDS gives string value (name of enum element, not numerical value)
var RequirementEnum = Object.freeze({
    "Unset": -1,
    "Mandatory": "Mandatory",
    "Optional": "Optional",
    "Conditional": "Conditional",
    "C1": "C1",
    "C2": "C2",
    "C2orC3": "C2orC3",
    "C3": "C3",
    "C4": "C4",
    "C5": "C5",
    "C6": "C6",
    "C7": "C7",
    "C8": "C8",
    "C9": "C9",
    "C10": "C10",
    "C11": "C11",
    "C12": "C12",
    "C13": "C13",
    "C14": "C14",
    "C15": "C15",
    "C16": "C16",
    "C17": "C17",
    "C18": "C18",
    "C19": "C19",
    "C20": "C20"

});

// Exports
if (typeof module !== 'undefined') {
    module.exports.createHeaderFile = createHeaderFile;
    module.exports.createCodeFile = createCodeFile;
    module.exports.enrich = enrich;
    module.exports.generateFiles = generateFiles;
}